<?php
namespace PrestaShop\Module\PolarisPrestaConnector\pol;

use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierClientWebConnectClientUpdateable;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierClientWebConnectFiltreClient;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\ProduitDetailTaille;
use PrestaShop\Module\PolarisPrestaConnector\BridgeAdapter;
use PrestaShop\Module\PolarisPrestaConnector\Models\Article;
use PrestaShop\Module\PolarisPrestaConnector\Models\AttributeRef;
use PrestaShop\Module\PolarisPrestaConnector\Models\Customer;
use PrestaShop\Module\PolarisPrestaConnector\Models\Discount;
use PrestaShop\Module\PolarisPrestaConnector\Models\DiscountPlan;
use PrestaShop\Module\PolarisPrestaConnector\SyncConfiguration;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\CatalogApi;
use PrestaShop\Module\PolarisPrestaConnector\Models\Product;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\ConfigApi;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\InfoConnect;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\Pager;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\StocksApi;
use PrestaShop\Module\PolarisPrestaConnector\Models\Stock;
use PrestaShop\Module\PolarisPrestaConnector\Models\Voucher;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\ClientsApi;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\VentesApi;

/**
 * Pont vers Polaris
 * 
 * Memo : utilisation de l'API Polaris possible ici 
 * $config = Configuration::getDefaultConfiguration()->setApiKey('X-API-Key', 'XX XXXX XXXX XXXX XXXX XXXX XXXX XXXX');
 * $config->getDefaultConfiguration()->setHost('https://XXX:14243/api/WebConnect');
 * $api = new CatalogApi(null, $config);
 */
class Bridge extends BridgeAdapter
{
    /**
     * Nombre de produits à récupérer en une fois
     */
    private const NB_PRODUCTS_AT_A_TIME = 100;

    /**
     * Nombre de clients à récupérer en une fois
     */
    private const NB_CUSTOMERS_AT_A_TIME = 100;

    /**
     * Nombre de stocks à récupérer en une fois
     */
    private const NB_STOCKS_AT_A_TIME = 10000;


    /**
     * Identifiants de liste de diffusion
     */
    public const CONST_DIFF_MAILING = 32768;    


    /**
     * URL de base de l'API
     */
    private string $baseUrl = "";

    /**
     * Informations de connexion
     */
    private ?InfoConnect $structure = null;

    /**
     * Cache de nom de grilles de tailles
     */
    private array $sizeGrids = [];

    /**
     * Api Polaris de produit
     */
    private ?CatalogApi $_catalogApi = null; 

    /**
     * Api Polaris de stock
     */
    private ?StocksApi $_stocksApi = null;

    /**
     * Api Polaris de ventes
     */
    private ?VentesApi $_ventesApi = null;

    /**
     * Api Polaris de clientèle
     */
    private ?ClientsApi $_clientsApi = null;

    /**
     * Dernière position de produits synchronisés
     */
    private string|null $lastProductPosition = null;

    /**
     * Dernière position de stocks synchronisés
     */
    private string|null $lastStockPosition = null;

    /**
     * Dernière position des promotions synchronisées
     */
    private string|null $lastDiscountPosition = null;

    /**
     * Dernière position des bons synchronisées
     */
    private string|null $lastVoucherPosition = null;

    /**
     * Dernière position des clients synchronisés
     */
    private string|null $lastCustomerPosition = null;

    /**
     * Plan de promotions
     */
    private ?array $discountsPlan = null;    

    /**
     * Constructeur
     */
    public function __construct(SyncConfiguration $config)
    {
        parent::__construct($config);        
    }

    /**
     * Obtient la configuration de connexion pour l'API
     */
    private function getApiConfig() : \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Configuration
    { 
        // On charge la configuration
        $loginCfg = $this->config->get(SyncConfiguration::CFG_LOGIN);

        if (!isset($loginCfg) || !isset($loginCfg['address']) || !isset($loginCfg['key']))
        {
            if (!empty($loginCfg))
            {
                $loginCfg = null;
                $this->config->set(SyncConfiguration::CFG_LOGIN, $loginCfg);  // reset !
            }

            throw new \Exception("Le module n'est pas associé au backoffice Polaris !");
        }

        $config = \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Configuration::getDefaultConfiguration();

        $config->setHost($loginCfg['address'] .  '/api/WebConnect');
        $config->setApiKey('X-API-Key', $loginCfg['key']);
        $config->setUserAgent('PolPolarisPrestaConnector/1.0; enum_as_str;');

        $this->baseUrl = $loginCfg['address'] . '/';

        return $config;
    }

    /**
     * Obtient l'API clients
     */
    private function getClientsApi() : ClientsApi
    {        
        if ($this->_clientsApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->_clientsApi = new ClientsApi(null, $apiCfg);
        }

        return $this->_clientsApi;
    }

    /**
     * Obtient l'API catalog
     */
    private function getCatalogApi() : CatalogApi
    {        
        if ($this->_catalogApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->_catalogApi = new CatalogApi(null, $apiCfg);
        }

        return $this->_catalogApi;
    }

    /**
     * Obtient l'API stocks
     */
    private function getStocksApi() : StocksApi
    {        
        if ($this->_stocksApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->_stocksApi = new StocksApi(null, $apiCfg);
        }

        return $this->_stocksApi;
    }

    /**
     * Obtient l'API ventes
     */
    private function getVentesApi() : VentesApi
    {        
        if ($this->_ventesApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->_ventesApi = new VentesApi(null, $apiCfg);
        }

        return $this->_ventesApi;
    }

    /**
     * Obtient les informations de l'API
     */
    private function getStructure() : InfoConnect
    {        
        if ($this->structure === null)
        {
            // Les infos n'ont pas encore été tirées, on s'en occupe !
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                return null;

            $infoApi = new ConfigApi(null, $apiCfg);
            $this->structure = $infoApi->configGetStruture();
            if (!$this->structure)
                throw new \Exception('Impossible de récupérer les informations de connexion');

            // BUG API Polaris : grilles de tailles
            $this->sizeGrids = [];
            foreach ($this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS) as $grid => $size)
                $this->sizeGrids[$grid] = $size;
        }

        return $this->structure;
    }

    /**
     * Renvoi l'identifiant du pont correspondant au backoffice géré
     */
    public function getId() : string
    {
        return 'Polaris';
    }

    /**
     * Renvoi les spécificités du pont sous forme de tableau associatif clé/valeur de type string
     * 
     * @return array<string,string>
     */
    public function getSpecifics() : array
    {
        //  La connexion avec votre backoffice LCV a été perdue. Merci de recommencer l'association.
        return [            
            "ADDRESS_LABEL" => "Quel est votre numéro de licence (il est de la forme pl-pX-XXXX-XXX)",
        ];
    }

    /**
     * Importe la configuration du backoffice via les API
     */
    public function pullEnvironment()
    {
        // Pour la suite, on va devoir tout charger, ça sera quand même bc plus rapide !
        if (!$this->config->hasBeenPreloaded())
            $this->config->preload();

        // Charger la structure
        $structure = $this->getStructure();

        // Fixation du magasin web
        if ($this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE) !== $structure->getMagasin()->getCode())
            $this->config->set(SyncConfiguration::CFG_WEB_STORE_CODE, $structure->getMagasin()->getCode());

        // Nom des catégories dynamiques
        $dyncat = $this->config->get(SyncConfiguration::CFG_DYN_CAT); 
        $dynattr = $this->config->get(SyncConfiguration::CFG_DYN_ATTR); 

        // Les catégories dynamiques
        foreach (['categorie1', 'categorie2', 'categorie3', 'categorie4'] as $cat)
        {
            $funcName = 'get' . ucfirst($cat);
            $catname = $structure->$funcName()->getNom();
            if (!isset($dyncat[$cat]) || $dyncat[$cat] != $catname)
                $this->config->setMap(SyncConfiguration::CFG_DYN_CAT, $cat, $catname);
        } 

        // Les attributs dynamiques
        foreach (['niveau1', 'niveau2', 'niveau3', 'niveau4'] as $attr)
        {
            $funcName = 'get' . ucfirst($attr);
            $attrname = $structure->$funcName()->getNom();
            if (!isset($dynattr[$cat]) || $dynattr[$cat] != $attrname)
             
            $this->config->setMap(SyncConfiguration::CFG_DYN_ATTR, $attr, $attrname);
        }

        $catalogApi = $this->getCatalogApi();

        // Tirage des valeurs des catégories dynamiques
        $empty = [ "*" => 0 ];
        $emptyName = [ "*" => "Tous les autres" ];
        foreach ([  'Categories1' => 'categorie1', 
                    'Categories2' => 'categorie2',
                    'Categories3' => 'categorie3',
                    'Categories4' => 'categorie4',
                    'Collections' => 'collection',
                    'Saisons' => 'saison'
                    ] as $func => $cat)
        {
            $funcName = 'catalogGet' . ucfirst($func);
            $mapName = SyncConfiguration::CFG_MAP_DYN_CAT . $cat;
            if ($this->config->get($mapName) === null)  /* auto création */
                $this->config->set($mapName, $empty);
            if ($this->config->get($mapName.'_name') === null)  /* auto création */
                $this->config->set($mapName.'_name', $emptyName);

            // On tire les catégorie            
            $pager = null;
            do
            {                
                $pageCategories = $catalogApi->$funcName(null, null, json_encode($pager));
                $pager = $pageCategories->getPagerNext();                

                foreach ($pageCategories->getItems() as $category)
                {
                    $code = $category->getNo();
                    $name = $category->getNom();
                    if ($this->config->getMap($mapName, $code) === null)
                        $this->config->setMap($mapName, $code, 0, true);

                    if ($this->config->getMap($mapName."_name", $category->getNo()) != $name)
                        $this->config->setMap($mapName."_name", $category->getNo(), $name);
                }
            }
            while ($pager != null);
        }

        // Tirage RFS
        $mapName = SyncConfiguration::CFG_MAP_DYN_CAT . 'rfs';
        if ($this->config->get($mapName) === null)  /* auto création */
            $this->config->set($mapName, $empty);
        if ($this->config->get($mapName.'_name') === null)  /* auto création */
            $this->config->set($mapName.'_name', $emptyName);

        // On tire les RFS
        $pager = null;
        do
        {                
            $pageCategories = $catalogApi->catalogGetClassifications(null, null, null, null, json_encode($pager));
            $pager = $pageCategories->getPagerNext();                

            foreach ($pageCategories->getItems() as $category)
            {
                $code = $category->getCode();
                $name = $category->getCode() . ' - ' . $category->getNom();
                if ($this->config->getMap($mapName, $code) === null)
                    $this->config->setMap($mapName, $code, 0, true);

                if ($this->config->getMap($mapName."_name", $code) != $name)
                    $this->config->setMap($mapName."_name", $code, $name);
            }
        }
        while ($pager != null);

        // Charger les grilles de tailles
        // BUG POLARIS : on ne peux pas tirer les grilles de tailles, alors on va les chercher en live (voir code dans nextProducts)
        
        // On tire les marques
        if ($this->config->get(SyncConfiguration::CFG_MAP_MANUFACTURER) === null)
            $this->config->set(SyncConfiguration::CFG_MAP_MANUFACTURER, $empty);

        $pager = null;
        do
        {
            $pageMarques = $catalogApi->catalogGetMarques(null, null, json_encode($pager));
            $pager = $pageMarques->getPagerNext();

            foreach ($pageMarques->getItems() as $marque)
            {
                if ($this->config->getMap(SyncConfiguration::CFG_MAP_MANUFACTURER, $marque->getNo()) === null)
                    $this->config->setMap(SyncConfiguration::CFG_MAP_MANUFACTURER, $marque->getNo(), 0, true);

                if ($this->config->getMap(SyncConfiguration::CFG_MAP_MANUFACTURER_NAME, $marque->getNo()) != $marque->getNom())
                    $this->config->setMap(SyncConfiguration::CFG_MAP_MANUFACTURER_NAME, $marque->getNo(), $marque->getNom());
            }
        }
        while ($pager != null);
    }

     /**
     * Retourne la liste exhaustive des options optionnelles supportées par ce pont
     * 
     * @return array<string> Liste des options optionnelles supportées     
     */
    public function getSupportedOptionalFeatures() : array
    {
        return [
            SyncConfiguration::CFG_IMPORT_CUSTOMERS,
        ];
    }

    /**
     * Révoque les configurations non supportées par le pont
     * 
     */
    public function getSupportedProductOptions() : array
    {
        return [
                SyncConfiguration::CFG_PRODUCT_SYNC_REF,
                SyncConfiguration::CFG_PRODUCT_SYNC_NAME,
                SyncConfiguration::CFG_PRODUCT_SYNC_MANUFACTURER,
                SyncConfiguration::CFG_PRODUCT_SYNC_CATEGORIES,
                SyncConfiguration::CFG_PRODUCT_SYNC_DESCRIPTION,
                SyncConfiguration::CFG_PRODUCT_SYNC_DIMENSIONS,
                SyncConfiguration::CFG_PRODUCT_SYNC_WEIGHT,
                //SyncConfiguration::CFG_PRODUCT_SYNC_ECOTAX,
                SyncConfiguration::CFG_PRODUCT_SYNC_TAXRATE,
                SyncConfiguration::CFG_PRODUCT_SYNC_EAN13,
                SyncConfiguration::CFG_PRODUCT_SYNC_LOCATION,
                SyncConfiguration::CFG_PRODUCT_SYNC_PHOTOS,
        ];
    }

    /**
     * Retourne la liste exhaustive des types de bons supportés par ce pont
     * 
     * @return array<string> Liste des type de bons supportés     
     */
    public function getSupportedVoucherOptions() : array
    {
        return [
            Voucher::AVOIR_VOUCHER,
            Voucher::GIFT_VOUCHER,
        ];
    }

    /**
     * Configuration par défaut de ce pont
     * 
     */
    public function alterDefaultCfg(array & $cfg)
    {
        $cfg[SyncConfiguration::CFG_DYN_CAT] = [
            'rfs' => 'RFS',
            'categorie1' => 'Catégorie 1',
            'categorie2' => 'Catégorie 2',
            'categorie3' => 'Catégorie 3',
            'categorie4' => 'Catégorie 4',            
            'collection' => 'Collection',
            'saison' => 'Saison', 
        ];

        $cfg[SyncConfiguration::CFG_DYN_ATTR] = [
            'niveau1' => 'Niveau 1',
            'niveau2' => 'Niveau 2',
            'niveau3' => 'Niveau 3',
            'niveau4' => 'Niveau 4',
        ];

        $cfg[SyncConfiguration::CFG_PR_ATTRIBUTES_MAPPING] = [
            SyncConfiguration::PRODUCT_ATTR_VARIANT => [
                'niveau1' => 0,
                'niveau2' => 0, // souvent la couleur
                'niveau3' => 0,
                'niveau4' => 0,
            ],
            SyncConfiguration::PRODUCT_ATTR_PRODUCT => [
                'niveau2' => 0, // souvent la couleur
            ],
            SyncConfiguration::PRODUCT_ATTR_FEATURE => [
                'collection' => 0, 
                'saison' => 0
            ],
        ];

        $cfg[SyncConfiguration::CFG_ORDER_PAY_MAPPAGE] = [
            '*' => 'CC',
            'Chèque' => 'CH',
            'Remise panier' => 'RE',
        ];
    }

    /**
     * Teste la connexion au backoffice avec les paramètres fournis
     */
    public function bind($args) : string|null
    {
        // Interressons nous à l'adresse qui est pour Polaris le numéro de licence
        // avec ce numéro, on va en déduire l'adresse de connexion
        
        if (!isset($args["address"]) || empty(trim($args["address"])))
            return "1 : le numéro de licence n'est pas défini !";

        $licenceId = strtolower(trim($args["address"]));

        // Le numéro de licence Polaris doit être de la forme "pl-pX-XXXX-XXX" où X est un chiffre ou "pl-vega-xxxx" où x est un chiffre, quelque soit le nombre de chiffres pour les deux formes
        if (!preg_match('/^pl-(p\d-)?\d+-\d+$/', $licenceId) && !preg_match('/^pl-vega-\d+$/', $licenceId))
            return "2 : le numéro de licence n'est pas valide !";        

        // On peut retrouver l'URL de connexion à partir du numéro de licence en se connectant au gestionnaire de licence polaris
        // avec l'URL suivante : https://extranet.vega-info.fr/manager/api/Public/polaris/XX/Url où XX est le numéro de licence (url escapé)
        // Cet endpoint nous renvoi un JSON avec l'adresse de connexion dans la propriété "url"
        $urlManager = 'https://extranet.vega-info.fr/manager/api/Public/polaris/' . urlencode($licenceId) . '/Url';
        $response = file_get_contents($urlManager);
        if ($response === false)
            return "3 : impossible de contacter le gestionnaire de licence Polaris. Merci de réessayer dans quelques instants. Si l'erreur se reproduit, merci de contacter le support technique de VEGA Informatique.";
        $url = json_decode($response)->url;

        // On vérifie maintenant que le pincode est correct. Il est rangé sous password
        if (!isset($args["pincode"]) || empty(trim($args["pincode"])))
            return "30 : renseignez maintenant le pincode de jumelage. Il est disponible dans votre backoffice Polaris."; 

        // On vérifie maintenant que le pincode est correct.
        // On va se connecter à l'API Polaris en POST avec les informations fournies pour transformer notre pincode en clé d'accès
        // Si on obtient un 401, c'est que le pincode est incorrect. Il nous faut donc le code de réponse HTTP du serveur
        // Pour cela, on va plutôt utiliser curl en GET   
        $ch = curl_init($url . '/api/Core/App/Binding?pinCode=' . urlencode($args["pincode"]));
        curl_setopt($ch, CURLOPT_HTTPGET, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true); // Permet de suivre les redirections
        curl_setopt($ch, CURLINFO_HEADER_OUT, true);
        $response = curl_exec($ch);
        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        if ($httpCode == 0)
            return "31 : impossible de contacter le backoffice Polaris. Vérifiez que vos ports sont ouverts sur votre box et que votre TLR est allumé !";

        if ($httpCode == 401)
            return "32 : le pincode de jumelage est incorrect. Merci de vérifier le code fourni.";

        if ($httpCode != 200)
            return "33 : une erreur inconnue (".$httpCode.") est survenue. Merci de contacter le support technique de VEGA Informatique.";

        // Si on arrive ici, on a toutes les informations nécessaires
        $response = json_decode($response, true);

        // sauver la config
        $loginCfg = [
            'address' => $url,
            'key' => $response["Guid"] . '/' . $response["Secret"],
        ];
        $this->config->set(SyncConfiguration::CFG_LOGIN, $loginCfg);        

        return parent::bind($args);
    }
    
    /**
     * Obtient l'URL d'une photo par sa clé
     */
    public function getPhotoURL(string $photo) : string
    {
        return $this->baseUrl . sprintf("api/Files/Media/Photo?noPhoto=%s&photoSize=original", $photo);
    }

    /**
     * Obtient le code magasin internet
     */
    public function getMagasinInternet() : string
    {
        $codeMagasin = $this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE);
        if (!$codeMagasin)
        {
            $this->pullEnvironment();   
            $codeMagasin = $this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE);
            if (!$codeMagasin)
                throw new \Exception('Erreur de configuration : le magasin web n\'est pas défini');
        }

        return $codeMagasin;
    }

    /**
     * Obtient les produits à synchroniser sous forme de tableau
     * 
     * @param string|null $productsRefsToSync Références des produits à tirer, ou null pour toutes
     * @return array<Product>
     */
    protected function & _pullProducts(?string $productsRefsToSync = null) : array
    {
        $catalogApi = $this->getCatalogApi();
        
        $products = [];
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastProductPosition)
        {            
            $this->setSyncPosition('products', $this->lastProductPosition);
            $this->lastProductPosition = null;
        }

        // Données de base de l'API BUG Grilles de taille
        $this->sizeGrids = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS);
        
        $codeMagasin = $this->getMagasinInternet();
        
        // Deux modes d'interogation :
        if ($productsRefsToSync)
        {
            // Mode référence
            // On extrait le no_modele (séparé par des pipes, c'est la première)
            $no_modele = explode('|', $productsRefsToSync)[0];
            // Nouveau filtre produit
            $filtreProduit = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\FiltreProduit();
            $filtreProduit->setNoModeles([$no_modele]);

            $pageProducts = $catalogApi->catalogGetProduits(json_encode($filtreProduit), null, null, null, 'true', 'true', 'true', 'true' );
        }
        else
        {
            // Mode pager            
            // Construction du pager
            $pager = new Pager();        
            $pager->setFrom($this->getSyncPosition('products') ?? '');
            $pager->setNb(self::NB_PRODUCTS_AT_A_TIME);
            $pager->setSort('+DerniereModification, +No');
            
            // endless pager ?
            $pager = json_encode($pager); // bug OpenAPI            

            $pageProducts = $catalogApi->catalogGetProduits(null, null, $pager, null, 'true', 'true', 'true', 'true' );            
        }

        if ($pageProducts === null || $pageProducts->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les produits');

        foreach ($pageProducts->getItems() as $source)
        {
            // Construction du produit
            $product = new Product($source->getNoModele(), $source->getCode());

            // Attributs du produit
            $product->productAttributes = [
                "categorie1" => new AttributeRef($source->getCategorie1()->getNo(), $source->getCategorie1()->getNom()),
                "categorie2" => new AttributeRef($source->getCategorie2()->getNo(), $source->getCategorie2()->getNom()),
                "categorie3" => new AttributeRef($source->getCategorie3()->getNo(), $source->getCategorie3()->getNom()),
                "categorie4" => new AttributeRef($source->getCategorie4()->getNo(), $source->getCategorie4()->getNom()),

                "niveau1" => new AttributeRef($source->getNiveau1()->getNo(), $source->getNiveau1()->getNom()),
                "niveau2" => new AttributeRef($source->getNiveau2()->getNo(), $source->getNiveau2()->getNom()),
                "niveau3" => new AttributeRef($source->getNiveau3()->getNo(), $source->getNiveau3()->getNom()),
                "niveau4" => new AttributeRef($source->getNiveau4()->getNo(), $source->getNiveau4()->getNom()),

                "collection" => new AttributeRef($source->getCollection()->getNo(), $source->getCollection()->getNom()),
                "saison" => new AttributeRef($source->getSaison()->getNo(), $source->getSaison()->getNom()),
                "rfs" => new AttributeRef($source->getClassification()->getCode(), $source->getClassification()->getNom()),
            ];

            // Grille de tailles
            $product->grilleTailles = $source->getGrilleTaille();

            // BUG API Polaris : on ne peux pas tirer les grilles de tailles, alors on va les chercher en live
            if (!array_key_exists($product->grilleTailles, $this->sizeGrids))
            {
                $this->sizeGrids = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS);
                $this->sizeGrids[$product->grilleTailles] = 0;
                $this->config->set(SyncConfiguration::CFG_MAP_SIZEGRIDS, $this->sizeGrids);

                $sizeGridsName = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS_NAME);
                $sizeGridsName[$product->grilleTailles] = $product->grilleTailles;
                $this->config->set(SyncConfiguration::CFG_MAP_SIZEGRIDS_NAME, $sizeGridsName);
            }

            // Taxe
            $product->taxrate = $source->getTauxTva();

            // Description
            $product->description = $source->getDescription();

            // Les dimensions / poids
            $product->weight = $source->getPoids();
            $volume = $source->getVolume();
            $product->depth = $volume[0] ?? null;
            $product->width = $volume[1] ?? null;
            $product->height = $volume[2] ?? null;

            $product->ecotax = $source->getEcoParticipationHt();

            // Les déclinaisons (tailles)            
            foreach ($source->getTailles() as $taille)
            {
                $refs = $taille->getRefs();
                if (empty($refs)) continue;

                $article = new Article($refs[0], $taille->getTaille());

                if (!empty($taille->getEans()))
                    $article->ean13 = $taille->getEans()[0];

                // Prix TTC
                foreach ($taille->getMagasins() as $store)
                    if ($store->getCodeMagasin() == $codeMagasin)
                        $article->priceTTC = $store->getPrixNormalTtc();

                // Zonage
                $article->warehouseLocation = $taille->getZonage();

                $product->articles[] = $article;
            }
            
            // Nom du produit
            $product->name = $source->getLibelle();

            // Marque
            $product->manufacturer = new AttributeRef($source->getMarque()->getNo(), $source->getMarque()->getNom());

            // Photos
            $product->photos = $source->getPhotos() ?? [];

            // Stock
            foreach ($source->getTailles() as $taille)
                $product->stocks[] = $this->buildStockFromTaille($taille);                    

            $products[] = $product;
        }

        // Fini
        if (!$productsRefsToSync)   // Pas de dernière position si on a filtré !
            $this->lastProductPosition = $pageProducts->getEndlessPager()->getFrom();
        return $products;
    }

    /**
     * Construit un stock à partir d'une response ProduitDetailTaille
     * 
     * @param ProduitDetailTaille $taille Taille à convertir
     * @return Stock Stock construit
     */
    private function buildStockFromTaille(ProduitDetailTaille $taille) : Stock
    {
        $codeMagasin = $this->getMagasinInternet();

        $prixTTC = null;
        $qties = [];
        foreach ($taille->getMagasins() as $store)
        {
            $qties[$store->getCodeMagasin()] = $store->getDispo();

            // Si on est dans le magasin Internet, on met le prix à jour
            if ($store->getCodeMagasin() == $codeMagasin)
                $prixTTC = $store->getPrixNormalTtc();
        }
        
        $stock = new Stock($taille->getRefs(), $qties);

        // Ean ?
        $stock->eans = $taille->getEans();

        // Prix ?
        $stock->prixTTC = $prixTTC;
    
        return $stock;
    }

    /**
     * Obtient les stocks à synchroniser sous forme de tableau
     * 
     * @return array<Stock>
     */
    protected function & _pullStocks(): array
    {
        $stocks = [];
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastStockPosition)
        {            
            $this->setSyncPosition('stocks', $this->lastStockPosition);
            $this->lastStockPosition = null;
        }
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('stocks') ?? '');
        $pager->setNb(self::NB_STOCKS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +No');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI        

        $pageStocks = $this->getStocksApi()->stocksGetStocks(null, null, $pager, null, 'false', 'false', 'false', 'false' );
        if ($pageStocks === null || $pageStocks->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les stocks');

        foreach ($pageStocks->getItems() as $source)
        {            
            foreach ($source->getTailles() as $taille)
                $stocks[] = $this->buildStockFromTaille($taille);
        }        

        // Fini
        $this->lastStockPosition = $pageStocks->getEndlessPager()->getFrom();
        return $stocks;
    }

    /**
     * Tire toutes les promotions depuis Polaris
     */
    private function pullDiscountPlans()
    {
        $this->discountsPlan = [];

        $pagerPlanPromos = null;
        do
        {
            $pagePlanPromos = $this->getStocksApi()->stocksGetPromotions(null, 'true', $pagerPlanPromos);
            if ($pagePlanPromos === null || $pagePlanPromos->getItems() === null)
                throw new \Exception('Erreur API Polaris : impossible de récupérer les plans de promotions');

            if (count($pagePlanPromos->getItems()) > 0)
            {
                foreach ($pagePlanPromos->getItems() as $planPromo)
                {
                    if (!$planPromo->getSurInternet())
                        continue; // Pas pour le web

                    $plan = new DiscountPlan($planPromo->getCode());
                    $plan->name = $planPromo->getNom();
                    $plan->dateStart = $planPromo->getDateDebut();
                    $plan->dateEnd = $planPromo->getDateFin();
                    
                    $this->discountsPlan[$plan->id] = $plan;
                }

                $pagerPlanPromos = $pagePlanPromos->getPagerNext();
            }
            else
                $pagerPlanPromos = null; // finito !
                
        } while ($pagerPlanPromos != null);
    }

    /**
     * Obtient les promotions à mettre à jour sous forme de tableau
     * 
     * @return array<DiscountPlan>
     */
    protected function & _pullDiscounts(): array
    {
        $discountsPlan = [];        
        
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastDiscountPosition)
        {
            $this->setSyncPosition('discounts', $this->lastDiscountPosition);
            $this->lastDiscountPosition = null;
        }
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('discounts') ?? '');
        $pager->setNb(self::NB_STOCKS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +Code');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI        

        $pagePromos = $this->getStocksApi()->stocksGetDetailsPromotions(null, 'false', null, null, $pager);
        if ($pagePromos === null || $pagePromos->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les promotions');

        foreach ($pagePromos->getItems() as $source)
        {
            // Ok, une promotion a été modifiée, chargeons les plans de promotions !
            if (!isset($this->discountsPlan))
                $this->pullDiscountPlans();

            if (count($discountsPlan) == 0)
                $discountsPlan = $this->discountsPlan;  // On copie les plans de promotions dans une variable locale
            
            $promoCode = $source->getCodePromotion();
            if (!isset($discountsPlan[$promoCode]))
                continue; // On ne gère plus cette promo, elle va être supprimée dans l'étape d'au dessus

            $promo = new Discount([$source->getRefProduit()]);
            $promo->eans = [$source->getEan()];
            $promo->discount = $source->getP();

            $discountsPlan[$promoCode]->discounts[] = $promo;
        }        

        // Fini
        $this->lastDiscountPosition = $pagePromos->getEndlessPager()->getFrom();
        
        // Pas de promotion pour l'instant
        if (count($discountsPlan) == 0)
        {
            // C'est fini, on peut libérer l'espace mémoire
            // pris par les plans promo
            $this->discountsPlan = null;
        }

        return $discountsPlan;
    }

    /**
     * Obtient les promotions à mettre à jour sous forme de tableau
     * 
     * @return array<Voucher>
     */
    protected function & _pullVouchers(): array
    {        
        $vouchers = [];
        
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastVoucherPosition)
        {
            $this->setSyncPosition('vouchers', $this->lastVoucherPosition);
            $this->lastVoucherPosition = null;
        }
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('vouchers') ?? '');
        $pager->setNb(self::NB_STOCKS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +Code');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI        

        $pageBons = $this->getClientsApi()->clientsGetBons(null, $pager);
        if ($pageBons === null || $pageBons->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les bons');

        foreach ($pageBons->getItems() as $source)
        {
            $bon = new Voucher(
                                $source->getCode(),
                                $source->getMontant() ?? 0,
                                $source->getPourcentage() ?? 0,
                            );

            // Nommage automatique
            $bon->name = $source->getLibelle();
            if (empty($bon->name))
            {
                switch ($source->getType())
                {
                    case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::CHEQUE_CADEAUX:
                        $bon->name = "Chèque cadeau";
                        break;
                    case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::BON_CADEAUX:
                        $bon->name = "Bon cadeau";
                        break;                    
                    case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::AVOIR:
                        $bon->name = "Avoir";
                        break;
                    case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::ARRHES:
                        $bon->name = "Arrhes";
                        break;
                    case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::ACOMPTE:
                        $bon->name = "Acompte";
                        break;            
                }
            }

            // Validité
            $bon->expirationDate = $source->getDateValidite();
            $bon->minimumAmount = $source->getSeuil();
            $bon->quantity = 1;
            if ($source->getAEteUtilise())
                $bon->quantity = 0;
            $bon->active = !$source->getQuarantaine();

            // Client ?
            $customer = $source->getClient();
            if ($customer && $customer->getNo() !== null && !empty($customer->getNo()))
            {
                $refsCustomer = $customer->getRefsExt() ?? null;
                if ($refsCustomer && count($refsCustomer))
                {
                    // Parce $refsCustomer[0] en int
                    $bon->customer_id = (int) $refsCustomer[0];
                }
                else
                {
                    $bon->customer_id = -1; /* client inexistant sur cette plateforme */
                }
            }

            switch ($source->getType())
            {
                case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::CHEQUE_CADEAUX:
                case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::BON_CADEAUX:                    
                    // Importation OK sous option
                    $bon->type = Voucher::GIFT_VOUCHER; // Cadeaux
                    break;                    
                case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::AVOIR:
                    // Importation OK sous option
                    $bon->type = Voucher::AVOIR_VOUCHER; // Avoir
                    break;
                case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::ARRHES:                
                case \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\TTypeBon::ACOMPTE:
                default:
                    // type de bon non importé
                    $bon->active = false;
                    break;
            }

            $vouchers[] = $bon;
        }        

        // Fini
        $this->lastVoucherPosition = $pageBons->getEndlessPager()->getFrom();        

        return $vouchers;
    }

    /**
     * Pousse les commandes indiquées vers le backoffice comme une seule et unique vente rattaché au client de la première
     * 
     * @param array<\Order> $psOrders Commandes à importer comme une seule et unique vente sous le backoffice
     * @return string|null null si la commande a été exportée, autrement justification de la non-exportation sans erreur
     */
    protected function _pushOrdersAsOneSale(array $psOrders) : string|null
    {
        // PS : $psOrders a déjà été vérifié

        // TODO : on charge la devise liée à la commande
        $clientsApi = $this->getClientsApi();

        // Commande de référence
        $psOrder = $psOrders[0];

        // Les ventes sous Polaris se font exclusivement dans une seule devise, 
        // on calcul donc un taux de conversion
        $cv_rate = 1.0 / $psOrder->conversion_rate;            

        // On détermine le tag d'export de la commande
        // On va déjà chercher l'année dans date_add qui est codé comme suit : "YYYY-MM-DD HH:MM:SS"
        $tag = substr($psOrder->date_add, 0, 4) . '/' . $psOrder->reference;

        // On construit la vente
        $vente = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierDebitWebConnectVenteUpdateable();
        $vente->setRefExt($tag);    // Idempotence
        $vente->setCodeCaisse(null);
        $vente->setCodeCaissier(null);
        $vente->setTypeVenteInternet('Site'); // 0 = Vente Internet
        $vente->setCodeMagasin($this->getMagasinInternet());
        $vente->setMemo(sprintf("Commande PrestaShop %s", $psOrder->reference));
        
        // On ajoute les produits
        $products = [];
        foreach ($this->getOrderProducts($psOrders) as $orderDetail)
        {                
            $product = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierDebitWebConnectDetailVente();

            // Calcul du SKU
            // La référence stockée dans la commande sinon l'ean13 du produit !
            $sku = $orderDetail["product_supplier_reference"];
            if (!$sku) $sku = $orderDetail["product_ean13"];
            $product->setSku($sku);
            $product->setQte($orderDetail["product_quantity"]);
            $product->setMontantTtc(round($orderDetail["total_price_tax_incl"] * $cv_rate, 2));

            // Remise ?
            if ($orderDetail["unit_price_tax_excl"] != $orderDetail["original_product_price"])
            {
                // Calcul de la remise
                $remise = ($orderDetail["original_product_price"] * $orderDetail["product_quantity"] * (100+$orderDetail["tax_rate"]))/100.0 - $orderDetail["total_price_tax_incl"];
                $product->setRemiseTtc(round($remise * $cv_rate, 2));
                $product->setTypeRemise('Promo'); // 0 = Normal ou 3 = Remise article
            }
            else
                $product->setTypeRemise('Normal'); // 0 = Normal ou 3 = Remise article

            $product->setTauxTva($orderDetail["tax_rate"]);
            
            $products[] = $product; // Ajout du produit
        }
        $vente->setDetails($products);

        // Calcul du montant total payé
        $totalPaid = 0;
        foreach ($psOrders as $psOrder)
            $totalPaid += $psOrder->total_paid;

        // On ajoute les paiements
        $psPayments = \OrderPayment::getByOrderReference($psOrder->reference);
        $paiements = [];
        $checkTotal = 0;
        $dateOrder = null;
        foreach ($psPayments as $payment)
        {
            $checkTotal += $payment->amount;            
            // paiement ?
            $pay_code = $this->mapPayment($payment->payment_method);
            if (!$pay_code || empty($pay_code))
                return sprintf("le mode de paiement '%s' n'est pas mappé", $payment->payment_method);   // Commande ignorée !
            $paiement = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierDebitWebConnectReglement();
            $paiement->setCode($pay_code);   // Carte de crédit par défaut            
            $paiement->setMontant(round($payment->amount * $cv_rate, 2));
            $paiement->setTransactionId($payment->transaction_id);

            if (!$dateOrder)
                $dateOrder = new \DateTime($payment->date_add);

            $paiements[] = $paiement;
        } 
        
        // Gestion des remises paniers
        $totalRemises = 0;
        $bons = [];
        foreach ($psOrders as $psOrder)
        {
            $usedVoucher = false;            

            $cartRules = $psOrder->getCartRules();
            foreach ($cartRules as $cartRule)
            {
                $rule = new \CartRule($cartRule['id_cart_rule']);
                if ($rule && $rule->id && !empty($rule->code))
                {
                    // Nous allons vérifier si ce voucher existe sur Polaris.
                    // Si oui, alors on l'ajoute aux vouchers, sinon, il sera traité comme une remise caisse
                    try
                    {
                        $polarisVoucher = $clientsApi->clientsGetBon($rule->code);
                        if ($polarisVoucher)
                        {
                            $usedVoucher = true;
                            $bons[] = $rule->code;
                        }
                    }
                    catch (\PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\ApiException $e)
                    {
                        if ($e->getCode() != 404)
                            throw $e; // Erreur inattendue
                    }
                }
            }

            if (!$usedVoucher)
                $totalRemises += $psOrder->total_discounts_tax_incl;
        }

        // Les remises paniers non gérées comme des bons ...
        if ($totalRemises > 0)
        {
            // Remise panier
            $code_remise = $this->mapPayment("Remise panier", false);
            if (!$code_remise || empty($code_remise))
                return sprintf("le mode de paiement '%s' n'est pas mappé", "Remise panier");   // Commande ignorée !

            $paiement = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierDebitWebConnectReglement();
            $paiement->setCode($code_remise);   // Carte de crédit par défaut            
            $paiement->setMontant(round($totalRemises * $cv_rate, 2));
    
            $paiements[] = $paiement;
        }

        // Bons !
        if (count($bons) > 0)
            $vente->setUtiliserBons($bons);

        // S'il y a une différence entre le montant payé et le montant de la commande, on ne l'importe pas !
        $vente->setReglements($paiements);
        if ($checkTotal != $totalPaid)
            throw new \Exception(sprintf('Erreur : le montant payé (%.2f) ne correspond pas au montant de la commande (%.2f) !', $checkTotal, $totalPaid));

        // Légalement, la date de la vente est la date du premier paiement !
        $vente->setDateVente($dateOrder);        

        if ($this->isCustomerValidForExport($psOrder->id_customer))
        {
            $psCustomer = new \Customer($psOrder->id_customer);
            if ($psCustomer && $psCustomer->id)
            {
                $client = new \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\MetierClientWebConnectClientMinimal();
                $client->setRefsExt([$psOrder->id_customer]);
                $client->setMail($psCustomer->email);
                $client->setNom($psCustomer->lastname);
                $client->setPrenom($psCustomer->firstname);                

                $vente->setClient($client);
            }                
        }
        
        // Puis on enregistre la vente dans Polaris
        try
        {            
            $this->getVentesApi()->ventesPostVente($vente, 'false');
        }
        catch (\PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\ApiException $e)
        {
            $msgBody = trim($e->getResponseBody());
            if (!$msgBody)
                $msgBody = $e->getMessage();
            // Si le message commence par un { et se fini par }, c'est un JSON
            if (substr($msgBody, 0, 1) == '{' && substr($msgBody, -1, 1) == '}')
            {
                $errMsg = json_decode($msgBody);
                if ($errMsg && isset($errMsg->Message))
                    throw new \Exception('Erreur API Polaris : impossible de créer la vente : '. $errMsg->Message);
            }            
            
            throw new \Exception('Erreur API Polaris : impossible de créer la vente : '. $msgBody);
        }

        return null;
    }

    /**
     * Obtient les clients à synchroniser sous forme de tableau
     * 
     * @return array<Customer>
     */
    protected function & _pullCustomers(): array
    {
        $customers = [];
        
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastCustomerPosition)
        {
            $this->setSyncPosition('customers', $this->lastCustomerPosition);
            $this->lastCustomerPosition = null;
        }
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('customers') ?? '');

        $pager->setNb(self::NB_CUSTOMERS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +No');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI        

        $filtre = new MetierClientWebConnectFiltreClient();
        $filtre->setExEmail(true); // Exclure les clients sans email

        $pageClients = $this->getClientsApi()->clientsGetClients(json_encode($filtre), $pager);
        if ($pageClients === null || $pageClients->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les clients');

        foreach ($pageClients->getItems() as $source)
        {
            $id = $source->getRefsExt();
            if ($id && count($id))
                $id = $id[count($id)-1]; // On prend le dernier ID
            else
                $id = 0;

            $customer = new Customer($id ?? 0, $source->getMail());
            $customer->firstname = $source->getPrenom();
            $customer->lastname = $source->getNom();
            $customer->civilite = $source->getIdent();
            $customer->newsletter = in_array(self::CONST_DIFF_MAILING, $source->getNoListeDiffusions());

            $customers[] = $customer;
        }        

        // Fini
        $this->lastCustomerPosition = $pageClients->getEndlessPager()->getFrom();        

        return $customers;
    }

    /**
     * Pousse le client indiqué vers le backoffice
     * 
     * @param \Customer $psCustomer Client à exporter
     * @return string|null null si le client a été exporté, autrement justification de la non-exportation sans erreur
     */
    protected function _pushCustomer(\Customer $psCustomer) : string|null
    {
        $client = new MetierClientWebConnectClientUpdateable();
        $client->setMail($psCustomer->email);
        $client->setNom($psCustomer->lastname);
        $client->setPrenom($psCustomer->firstname);
        $client->setRefsExt([$psCustomer->id]);
        $civilite = $this->getGenderFromPsGenderId($psCustomer->id_gender);
        if ($civilite)
            $client->setIdent($civilite); 

        if ($psCustomer->newsletter)
            $client->setAjouterAListeDiffusions([self::CONST_DIFF_MAILING]);
        else
            $client->setRetirerDeListeDiffusions([self::CONST_DIFF_MAILING]);

        // On envoi le client via l'API
        try
        {            
            $result = $this->getClientsApi()->clientsPostClient($client);
            if ($result === null)
                return "Erreur API Polaris : impossible de créer le client";
        }
        catch (\PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\ApiException $e)
        {
            $msgBody = trim($e->getResponseBody());
            if (!$msgBody)
                $msgBody = $e->getMessage();
            
            // Si le message commence par un { et se fini par }, c'est un JSON
            if (substr($msgBody, 0, 1) == '{' && substr($msgBody, -1, 1) == '}')
            {
                $errMsg = json_decode($msgBody);
                if ($errMsg && isset($errMsg->Message))
                    throw new \Exception('Erreur API Polaris : impossible de créer/modifier le client : '. $errMsg->Message);
            }            
            
            throw new \Exception('Erreur API Polaris : impossible de créer/modifier le client : '. $msgBody);
        }

        return null;
    }
}
